--- Redirects chat command to AceConfig processor or show's the sdkp options panel 
-- if there is no command specified.
-- @param input The command issued.
function sdkp:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub("AceConfigCmd-3.0").HandleCommand(sdkp, "sdkp", "sdkp", input)
    end
end

--- Purge command.  Clears all snapshot, loot, and bid data.
-- @input info AceConfig context object.
function sdkp:Purge(info)
    self.db.profile.snapshots = {};
    self.db.profile.waitlist = {};
    self.db.profile.waitlistData = {};
    self.db.profile.banlist = {};
    self.db.profile.raidPresence = {};
    self.db.profile.bids = {};
    while self.db.profile.loot[1] do 
   		-- if we replace the table with a new one, the old reference is retained in the scrollingtable widget.
    	table.remove(self.db.profile.loot); -- remove the loot
    end
   	while SDKP_FILTERS_LIST[3] do 
   		SDKP_FILTERS[SDKP_FILTERS_LIST[#SDKP_FILTERS_LIST]] = nil;
   		table.remove(SDKP_FILTERS_LIST); -- remove the last filter
   	end
   	self.newui.filtercontrol:Fire("OnValueChanged", 1);
    self:Print("Snapshot and loot data purged; and presence list reset.");
	self:RefreshScrollTables();
end

--- Snapshot command.  Takes a snapshot of everyone in the raid and fires an attendance check to the 
-- guild to allow them to register themselves in the snapshot. Also announces loot.  
-- This event is called automatically when opening loot for a boss identified from LibBabble-Boss.
-- @param info AceConfig context object.
-- @param notes snapshot notes, if empty considered a manual snapshot.
-- @param loot the loot table of the boss passed when opening the boss' loot window.
-- @param fForceFullSnap force the mod to collect a brand new snapshot of everyone in the raid instead of carrying over att data from the start.
function sdkp:CmdSnapshot(info, notes, loot, fForceFullSnap)	
	self:Debug("Taking a snapshot...", info, notes, loot, fForceFullSnap);
	local snapshots = self.db.profile.snapshots;
	local waitlist = self.db.profile.waitlist;
	notes = strtrim(notes or "");
	
	if not sdkp:IsTrackLoot() then 
		sdkp:Debug("Loot tracking is not on, skipping snapshot.");
		return;
	elseif notes == "" then -- manual snapshot
		-- Going to try my best to figure out what boss it is here.  Loot is probably a chest or something
   		-- Look up the subzone in the zone to encounter table
   		local subzone = SDKP_ZONE_TO_ENCOUNTER[ GetSubZoneText()];
	   	if subzone then 
	   		notes = subzone;
	   	else
	   		notes = GetRealZoneText() .. " - Manual Snapshot";
	   	end
	end
	
	local thissnap = snapshots[#snapshots];
	if not(thissnap) or thissnap.name ~= notes then 
		-- take the snapshot
		table.insert(snapshots, {
			["name"] = notes, -- [1]
			["playertable"] = {},
			["timestamp"] = time(),
			["zone"] = GetRealZoneText(),
		});
		local snapindex = #snapshots;
		thissnap = snapshots[snapindex];
		self:Debug("Snapshot inserted into snapshots table.");
		
		local nRaidMembers = GetNumRaidMembers();
		self.takingsnapshot = true;
		if snapindex == 1 or not self:IsStartAtt() or fForceFullSnap then
			for raidIndex = 1, nRaidMembers do
				local raidMember = GetRaidRosterInfo(raidIndex);
				local main = self:GetAltInfo(raidMember);
				if main then 
					raidMember = main;	
				end
				self:CmdHandleManualSnapshot(raidMember, "add", thissnap);
				self:CmdHandleManualWaitlist(raidMember, "remove");
			end
		else
			thissnap["playertable"] = nil;
		end
		self.takingsnapshot = nil;
		
		self:Debug("Shout out to the homies...", self:IsAnnounceLoot(), loot, self:GetAnnounceWhere());
		self:SendChatMessage("["..notes.."] BOSS DOWN!!!",  self:GetAnnounceWhere());
		if self:IsAnnounceLoot() and loot then
			local count = 1;
			local message = ""; 
			for item in pairs(loot) do 
				message = message .. item;
				count = count + 1;
				if count == 3 then 
					self:SendChatMessage(message, self:GetAnnounceWhere());
					message = "";
					count = 0;
				end
			end
			if message ~= "" then 
				self:SendChatMessage(message, self:GetAnnounceWhere());
			end
		end
		
		-- self:IsStartAtt() is true for a new system where attendance is checked at the start of the raid.
		-- so only announce an attendance check each kill if we aren't using this new system
		if self:IsAttCheck() and not self:IsStartAtt() then 
			self:SendGuild("ATTENDANCE CHECK! Response needed from NON raid members within ".. string.format("%.2f", self:GetAttCheckLen()/60) .. " min.");
			self:SendGuild("In order to get credit for the night, you must [  /w " .. UnitName("player") .. " !here ]");
			self:ScheduleTimer(function() 
				self.checkingatt = false;
				self:SendGuild("ATTENDANCE CHECK HAS ENDED!  That is all...");
			end, self:GetAttCheckLen(), nil);
			self.checkingatt = true;
		end
		
		if self:IsWaitlist() then
			local addtosnap = self:GetLastFullSnapshot(snapindex);
			-- add the waitlist to the snapshot
			for raidWaiter,isThere in pairs(waitlist) do
				self:CmdHandleManualSnapshot(raidWaiter, "add", addtosnap);
			end
		end
		
		self:AddRaidFilter(notes, self:GetLastFullSnapshot().playertable);
	end
	
	self:RefreshScrollTables();
end

-- print debug mode handlers
function sdkp:IsDebug(info)
    return self.db.profile.debug;
end
function sdkp:ToggleDebug(info, value)
    self.db.profile.debug = value;
    SDKP_RAID_ZONE["Scarlet Monastery"] = self.db.profile.debug; -- easy to test on sm bosses, close to zone.
end

-- Loot handlers
function sdkp:IsAnnounceLoot(info, key)
    return self.db.profile.announceloot;
end
function sdkp:ToggleAnnounceLoot(info, value)
	self.db.profile.announceloot = value;
end

function sdkp:GetAnnounceWhere(info, ...)
    return self.db.profile.announcewhere;
end
function sdkp:SetAnnounceWhere(info, key)
	self.db.profile.announcewhere = key;
end

function sdkp:IsTrackLoot(info)
    return self.db.profile.trackloot;
end
function sdkp:ToggleTrackLoot(info, value)
    self.db.profile.trackloot = value;
    if value then
    	self:RegisterEvent("READY_CHECK");
    else
    	self:UnregisterEvent("READY_CHECK");
    end
end

function sdkp:GetBidChannel(info, ...)
    return self.db.profile.bidchannel;
end
function sdkp:SetBidChannel(info, key)
	self.db.profile.bidchannel = key;
end

function sdkp:GetLootFilter(info, key)
    return self.db.profile.lootfilter[SDKP_COLORS[key]];
end
function sdkp:SetLootFilter(info, key, value)
	self.db.profile.lootfilter[SDKP_COLORS[key]] = value;
end

-- Waitlist handlers
function sdkp:IsWaitlist(info)
    return self.db.profile.checkwaitlist;
end
function sdkp:ToggleWaitlist(info, value)
    self.db.profile.checkwaitlist = value;
end

-- Timers
function sdkp:GetAttCheckLen(info)
    return self.db.profile.attchecklen;
end
function sdkp:SetAttCheckLen(info, value)
    self.db.profile.attchecklen = value;
end
function sdkp:GetSanctionLen(info)
    return self.db.profile.sanctionlen;
end
function sdkp:SetSanctionLen(info, value)
    self.db.profile.sanctionlen = value;
end
function sdkp:GetAttStartGraceLen(info)
    return self.db.profile.attstartgracelen;
end
function sdkp:SetAttStartGraceLen(info, value)
    self.db.profile.attstartgracelen = value;
end


-- Dkp system handlers
function sdkp:IsAttCheck(info)
    return self.db.profile.checkattendance;
end
function sdkp:ToggleAttCheck(info, value)
    self.db.profile.checkattendance = value;
end
function sdkp:IsStartAtt(info)
    return self.db.profile.attendancestart;
end
function sdkp:ToggleStartAtt(info, value)
    self.db.profile.attendancestart = value;
end
function sdkp:GetTaxThreshold(info)
    return self.db.profile.taxthreshold;
end
function sdkp:SetTaxThreshold(info, value)
    self.db.profile.taxthreshold = value;
end
function sdkp:GetTaxRate(info)
    return self.db.profile.taxrate;
end
function sdkp:SetTaxRate(info, value)
    self.db.profile.taxrate = value;
end
function sdkp:IsRelationalDkp(info)
    return self.db.profile.relationaldkp;
end
function sdkp:ToggleRelationalDkp(info, value)
    self.db.profile.relationaldkp = value;
	self:RefreshScrollTables();
end
function sdkp:GetInitialGearValue(info)
    return self.db.profile.initialgearvalue;
end
function sdkp:SetInitialGearValue(info, value)
	self.db.profile.initialgearvalue = value;
	self:RefreshScrollTables();
end
function sdkp:IsShroudDkp(info)
    return self.db.profile.shrouddkp;
end
function sdkp:ToggleShroudDkp(info, value)
    self.db.profile.shrouddkp = value;
	self:RefreshScrollTables();
end

-- Ignore loot handlers
function sdkp:CmdIgnoreLoot(info, link)
	if link == "" then 
		self:Print("|cffffff78Usage:|r /sdkp ignoreloot [Item]    |cffffff78Example:|r /sdkp il |cff0070dd|Hitem:32428:0:0:0:0:0:0:0|h[Heart of Darkness]|h|r");
		self:Print(self:ListIgnoreLoot());
		return;
	end
	local _, _, item = string.find(link, "|c%x+|Hitem:%d+[:%d%-]+|h%[(.-)%]|h|r");
	if not item then 
		item = link;
	end		 
	if self.db.profile.ignoreloot[item] then
		self.db.profile.ignoreloot[item] = nil;
		self:Print("No longer ignoring " .. link);
	else
		self.db.profile.ignoreloot[item] = true;
		self:Print("Now ignoring " .. link);
	end
end
function sdkp:IsIgnoreLoot(item)
	return self.db.profile.ignoreloot[item] == true;
end
function sdkp:ListIgnoreLoot()
	local count = 0;
	for key, value in pairs(self.db.profile.ignoreloot) do
		count = count + 1;
	end
	return "Currently ignoring " .. count .. " items.";
end

-- Assign loot handlers
function sdkp:CmdAssignLoot(info, args)
	if args == "" then 
		self:Print("|cffffff78Usage:|r /sdkp assignloot [Item] player (reason/note)");
		self:Print("|cffffff78Example:|r /sdkp al |cff0070dd|Hitem:32428:0:0:0:0:0:0:0|h[Heart of Darkness]|h|r Sachmo");
		self:Print("|cffffff78Example:|r /sdkp al |cff0070dd|Hitem:32428:0:0:0:0:0:0:0|h[Heart of Darkness]|h|r Sachmo because he is awesome!");
		return;
	end
	self:Debug("sdkp:CmdAssignLoot", info, args);
	string.gsub(args, "%s*(|c%x+|Hitem:%d+[:%d%-]+|h%[.-%]|h|r)%s+([^%s]+)%s*(.*)", function(link, player, note)
		if not note or note == "" then 
			note = "Manually assigned";
		end
		self:Debug("sdkp:CmdAssignLoot GSUB", link, player, note);
		self:CHAT_MSG_LOOT(note, player .. " receives loot: " .. link .. ".");
		return false;
	end);
end

-- Manual attendance/dkp snapshots handler
function sdkp:CmdHandleManualSnapshot(player, command, snapshot)
	if not player or strtrim(player) == "" then
		return;
	end
	player = strtrim(player);
	local main = self:GetAltInfo(player);
	if main then 
		player = main;
	end
	local message;
	
	if command == "add" then
		if snapshot.playertable[player] then
			message = player .. " was already in the snapshot.";
		else
			snapshot.playertable[player] = true;
			message = player .. " added to the snapshot.";
		end
	elseif command == "remove" then 
		if snapshot.playertable[player] then
			snapshot.playertable[player] = nil;
			message = player .. " has been removed from the snapshot.";
		else
			message = player .. " was not in the snapshot.";
		end
	end
	if not self.takingsnapshot then 
		self:Print(message);
	end
 
	self:RefreshScrollTables();
end

-- Manual waitlist handler
function sdkp:CmdHandleManualWaitlist(player, command)
	local snapshots = self.db.profile.snapshots;
	local waitlist = self.db.profile.waitlist;
	local banlist = self.db.profile.banlist; 
	
	if not player or strtrim(player) == "" then
		return;
	end
	player = strtrim(player);
	-- Get the main of the whispering character
	local main = self:GetAltInfo(player);
	if main then 
		player = main;
	end
	-- Get the player info from dkpData to ensure proper formatting
	local dkpDataExists = self:getPlayerData(player);
	if dkpDataExists then
		player = dkpDataExists.cols[1].value;
	end 
	local message;
	
	-- Check for player on waitlist 
	if waitlist[player] then
		if command == "add" then
			message = player .. " is already on the waitlist.";
		elseif command == "remove" or command == "strip" then 
			waitlist[player] = nil;
			message = player .. " has been removed from the waitlist.";
		end
		if not(command == "strip" or command == "ban") then
			if not self.takingsnapshot then 
				self:Print(message);
			end
		 	return;
		end
	end
	
	if command == "add" then 
		if not banlist[player] then -- check is player is banned first
			waitlist[player] = true;
			message = player .. " added to the waitlist.";
		else
			message = player .. " is banned for the night.";
		end
	elseif command ~= "strip" then 
		message = player .. " was not on the waitlist.";
	end
	
	if command == "strip" then
 		-- strip him from all saved snapshots
		for _, snapshot in ipairs(snapshots) do
			if snapshot.playertable then 
				sdkp:CmdHandleManualSnapshot(player, "remove", snapshot);
			end
		end
		message = player .. " has been stripped from all saved snapshots.";
 	end
 	
 	self:Debug("Player:command " .. player .. ":" .. command);
 	if command == "ban" then
 		if banlist[player] then
 			message = plalyer .. " is already on the banlist.";
 		else
	 		banlist[player] = true;
	 		message = player .. " has been added to the banlist.";
	 	end
	end
 	
 	if command == "unban" then
 		if banlist[player] then
 			banlist[player] = nil;
 			message = player .. " has been removed from the banlist.";
 		else
 			message = player .. " was not on the banlist.";
 		end
 	end
 	
	if not self.takingsnapshot then 
		self:Print(message);
	end
 
 	self:RefreshScrollTables();
end

-- Start the raid. Announce that the raid has started
function sdkp:CmdStartRaid(info)
	local gracePeriod = self:GetAttStartGraceLen();
	
	local notes = GetRealZoneText() .. ":Start";
	self:CmdSnapshot(info, notes, nil, true);
	
	self.checkingatt = true;	-- Allow !here whispers
	
	self:Debug("Belt out to the brothas...");
	-- self:SendGuild("RAID START! To earn DKP you must be on the attendance snapshot.");
	-- self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !here ]");
	
	-- Spam the guild every 5 minutes during the grace period to make sure people see it
	for i0 = 300, gracePeriod, 300 do
		self:ScheduleTimer(function()
			self:SendGuild("RAID START! To earn DKP you must be on the attendance snapshot. " .. i0/60 .. " minutes left to add yourself to the snapshot.");
			self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !here ]");
		end, gracePeriod-i0, nil);
	end
	-- One more warning at 1 minute remaining
	local i0 = 60;
	self:ScheduleTimer(function()
		self:SendGuild("RAID START! To earn DKP you must be on the attendance snapshot. " .. i0/60 .. " minute left to add yourself to the snapshot.");
		self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !here ]");
	end, gracePeriod-i0, nil);
	-- Missed the boat
	self:ScheduleTimer(function()
		self.checkingatt = false;	-- Close !here whispers
		self:SendGuild("RAID START! Grace period elapsed, that is all...");
	end, gracePeriod, nil);
	
	self:SendMessage("SDKP_LOOT_ASSIGNED", UnitName("player"), "Fakery", 0, "Fakery");
end

-- Stop the raid. Announce that the raid has ended
function sdkp:CmdStopRaid(info)
	local gracePeriod = self:GetAttStartGraceLen() + 5*60;
	
	local notes = GetRealZoneText() .. ":End";
	self:CmdSnapshot(info, notes, nil, true);
	
	self.checkingatt = true;	-- Allow !here whispers
	
	self:Debug("Belt out to the brothas...");
	-- Spam the guild every 5 minutes during the grace period to make sure people see it
	for i0 = 300, gracePeriod, 300 do
		self:ScheduleTimer(function()
			self:SendGuild("RAID END! Raiding is done for the night. " .. i0/60 .. " minutes left to add yourself to closing snapshot for bonus DKP.");
			self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !here ]");
		end, gracePeriod-i0, nil);
	end
	-- One more warning at 1 minute remaining
	local i0 = 60;
	self:ScheduleTimer(function()
		self:SendGuild("RAID END! Raiding is done for the night. " .. i0/60 .. " minute left to add yourself to closing snapshot for bonus DKP.");
		self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !here ]");
	end, gracePeriod-i0, nil);
	-- Missed the boat
	self:ScheduleTimer(function()			
		self.checkingatt = false;	-- Close !here whispers
		self:SendGuild("RAID END! Grace period elapsed, that is all...");
	end, gracePeriod, nil);
	
	self:SendMessage("SDKP_LOOT_ASSIGNED", UnitName("player"), "Fakery", 0, "Fakery");
end

-- Start tracking the waitlist
-- enable adding the waitlist to snapshots and announce that the raid has started
function sdkp:CmdStartWaitlist(info)
	local spamInterval = 1800;
	
	self:Debug("Belt out to the brothas...");
	if not self:IsWaitlist() then 
		self.db.profile.waitlist = {}; -- clear the waitlist
			
		self:SendGuild("RAID START! In order to get credit for boss kills, you must be on the waitlist.");
		self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !waitlist add ]");
	end
	
	-- Start a repeating timer to remind people a raid is in progress
	if not self.spamTimer then
		self.spamTimer = self:ScheduleRepeatingTimer(function()
				self:SendGuild("RAID IN PROGRESS! In order to get credit for boss kills, you must be on the waitlist.");
				self:SendGuild("Send a tell to be added: [  /w " .. UnitName("player") .. " !waitlist add ]");
			end, spamInterval, nil);
		self:Debug("spamTimer " .. self.spamTimer);
	end
	
	self:ToggleWaitlist(_, true);
end

-- Stop tracking the waitlist
-- disable adding the waitlist to snapshots, and announce the raid has ended
-- anyone still on the waitlist after the announced time has expired will be stripped from all saved snapshots
function sdkp:CmdStopWaitlist(info)
	local waitlist = self.db.profile.waitlist;
	local gracePeriod = 300;
	
	self:Debug("Belt out to the brothas...");
	if self:IsWaitlist() then 
		-- Stop the repeating spam timer
		self:CancelTimer(self.spamTimer)
		
		-- Remove all current raid members from the waitlist
		self.takingsnapshot = true; -- Set taking snapshot to true to suppress message spam
		local nRaidMembers = GetNumRaidMembers();
		for raidIndex = 1, nRaidMembers do
			local raidMember = GetRaidRosterInfo(raidIndex);
			local main = self:GetAltInfo(raidMember);
			if main then 
				raidMember = main;	
			end
			self:CmdHandleManualWaitlist(raidMember, "remove");
		end
		self.takingsnapshot = nil;
		
		-- Spam the guild once a minute during the grace period to make sure people see it
		for i0 = 60, gracePeriod, 60 do
			self:ScheduleTimer(function()
				self:SendGuild("RAID END! Raiding is done for the night. Remove yourself from the waitlist in the next " .. i0/60 .. " minutes or you will be removed from all boss snapshots.");
				self:SendGuild("Send a tell to be removed: [  /w " .. UnitName("player") .. " !waitlist remove ]");
			end, gracePeriod-i0, nil);
		end
		self:ScheduleTimer(function()			
			-- add stragglers to the ban list
			for raidWaiter,isThere in pairs(waitlist) do
				self:Debug(raidWaiter .. " getting banned.");
				self:CmdHandleManualWaitlist(raidWaiter, "ban");
			end			
			self:SendGuild("RAID END! Grace period elapsed, that is all...");
		end, gracePeriod, nil);
		self:ToggleWaitlist(_, false);
	end
end

-- Convert the waitlist to a string for output
function sdkp:CmdWaitlistString(waitlist)
	local message, i = {}, 1;
	for raidWaiter,isThere in pairs(waitlist) do
		message[i] = raidWaiter;
		i = i + 1;
	end
	return table.concat(message, ", ");
end
		
-- Manually assign major/minor noobery to a player
function sdkp:CmdNoobery(type, args)
	if not args or strtrim(args) == "" then
		return;
	end
	local _, _, player, note = args:find("([^%s]+)%s*(.*)");
	if not note or note == "" then
		note = "Just Because...";
	end
		
	local which, price = "Minor Noobery", 1;
	if type == "major" then 
		which, price = "Major Noobery", 3;
	end
	local message = player .. " receives loot: " .. which .. ". (" .. note .. ")";
	self:SendOfficer(message);
	self:SendRaid(message);
	
	sdkp:SendMessage("SDKP_LOOT_ASSIGNED", player, which, price, note);
end

-- Export handler
function sdkp:CmdExport(info)
	local banlist = self.db.profile.banlist;
	if not self.db.profile.snapshots or #self.db.profile.snapshots < 1 then 
		self:Print("No snapshot data!");
		return;
	end

	local ts = self.db.profile.snapshots[1].timestamp;
	local time = date("%m/%d/%y %H:%M:%S", ts);
	
	local exportdata = "<RaidInfo><key>"..time.."</key><start>"..time.."</start><end>"..time.."</end><zone></zone>";
	exportdata = exportdata.."<BossKills></BossKills><note>" .. self.db.profile.snapshots[1].zone .. "</note>";
	
	local count, join, leave = 1, "", "";	
	local function fnFormatJoinLeave(previous, snap, num)
		local joints = ts + (10 * (num - 2));
		local leavets = ts + (10 * (num - 1)) - 1;
		if previous == snap then 
			leavets = joints + 50;
		end
		local jointime = date("%m/%d/%y %H:%M:%S", joints);
		local leavetime = date("%m/%d/%y %H:%M:%S", leavets);
		
		for player in pairs(previous.playertable) do 
			if not banlist[player] then
				join = join.."<key"..count.."><player>"..player.."</player><time>"..jointime.."</time></key"..count..">";
				leave = leave.."<key"..count.."><player>"..player.."</player><time>"..leavetime.."</time></key"..count..">";
				count = count + 1;
			end
		end	
	end
	
	local snapshots = self.db.profile.snapshots;
	for i, snap in ipairs() do 
		if snap.playerlist then 
			if i == 1 and #snapshots == 1 then 
				fnFormatJoinLeave(snap, snap, i);
			end
			
			if i ~= 1 then 
				local previous = snapshots[i - 1];
				fnFormatJoinLeave(previous, snap, i);
			end 
		end
	end
	fnFormatJoinLeave(snapshots[#snapshots], snapshots[#snapshots], #snapshots + 1);
	
	exportdata = exportdata.."<Join>"..join.."</Join><Leave>"..leave.."</Leave><Loot>";
	
	local i, j = #self.db.profile.loot, #snapshots;
	local boss;
	while i >= 1 do
		local itemdata = self.db.profile.loot[i];
		
		local name, id = itemdata.item, itemdata.id;
		if name:find("Fakery") then 
			name = "";
			id = "";
		end
		exportdata = exportdata.."<key"..i..">";
		exportdata = exportdata.."<ItemName>"..name.."</ItemName>";
		exportdata = exportdata.."<ItemID>"..id.."</ItemID>";
		exportdata = exportdata.."<Icon></Icon><Class></Class><SubClass></SubClass>";
		exportdata = exportdata.."<Color>"..itemdata.color.."</Color>";
		exportdata = exportdata.."<Count>1</Count>";
		exportdata = exportdata.."<Player>"..itemdata.player.."</Player>";
		
		while j >= 1 do
			if itemdata.timestamp >= snapshots[j].timestamp then 
				time = date("%m/%d/%y %H:%M:%S", ts + (10 * (j - 1)) + 1);
				boss = snapshots[j].name;
 				break;
			end		
			j = j - 1;	
		end
		if j == 0 then 
			time = date("%m/%d/%y %H:%M:%S", ts + 1);
			boss = snapshots[1].name;
		end
		
		exportdata = exportdata.."<Time>"..time.."</Time>";
		exportdata = exportdata.."<Note><![CDATA[ - Boss: "..boss.." - "..itemdata.price.." DKP]]></Note>";
		exportdata = exportdata.."</key"..i..">";
		i = i - 1;
	end
	exportdata = exportdata.."</Loot></RaidInfo>";
	
	self:EnqueueDialog(
		"CTRA DKP Export",
		"Ok", 
		"Cancel", 
		function()end,
		function()end,
		exportdata
	); 
	self:ConfirmPopup();
end

-- sync handler
function sdkp:CmdSync(info, args)
	if not args or strtrim(args) == "" then
		self:Print("Who to sync with?");
		return;
	end
	local _, _, who = args:find("([^%s]+)");
	self.requestsync = who;
	self:SendCommMessage(self.name, self:Serialize("requestsync"), "WHISPER", who, "BULK");
end

-- bids
function sdkp:StartBids(context, event, itemedit)
	local item = itemedit:GetText();
	local _, _, itemid = item:find("item:([^:]+):");
	itemid = tonumber(itemid);
	
	if context.text:GetText() == "Start Bids" then 
		self.BIDDING = true;
		context:SetText("Stop Bids");
		if sdkp:IsShroudDkp() then
			self:SendRaidWarning("Bidding started for " .. item .. " ( submit a bid from 1 to 50 )");
		else
			self:SendRaidWarning("Bidding started for " .. item .. " ( full = " .. self:GetDkpForItem(item) .. " )");
		end
		self:SendRaidWarning("To bid: '/w " .. UnitName("player") .. " !bid X', where X is your bid.");
				
		local classtable = nil;
		local PT = LibStub("LibPeriodicTable-3.1");
		local settable = PT:GetSetTable("Misc.Currency.Tier Sets.Class");
		for i, classset in ipairs(settable) do 
			if classset[itemid] then 
				if not classtable then 
					classtable = {}; 
				end
				local _, _, class = classset.set:upper():find("([^.]+)$");
				classtable[RAID_CLASS_COLORS[class]] = true;
			end
		end
		
		if classtable then 
			self:Print("Gear token found.  Setting filter to those who can hold it.  Clearing bids will restore old filter.");
			self.LASTFILTER = self.BIDSCROLLFRAME.Filter;
			self.BIDSCROLLFRAME:SetFilter(function (self, row)
				local ret = false;
				if SDKP_FILTERS[SDKP_FILTERS_LIST[2]](self, row) and classtable[row.cols[1].color]  then 
					ret = true;
				end
				return ret;		
			end);
		end
		
		local cols = self.BIDSCROLLFRAME.cols;
		for i, col in ipairs(cols) do
			if col.name == "Bid" then
				col.sort = "asc";
			else
				col.sort = nil;
			end 
		end
		self.BIDSCROLLFRAME:SortData();
	else
		self.BIDDING = false;
		context:SetText("Start Bids");
		self:SendRaidWarning("Bidding has ended for " .. item .. ".");
	end
end

function sdkp:ClearBids(context, event, itemedit)
	itemedit:SetText("");
	self.db.profile.bids = {};
	
	if self.LASTFILTER and self.BIDSCROLLFRAME.Filter ~= self.LASTFILTER then 
		self.BIDSCROLLFRAME:SetFilter(self.LASTFILTER);
	end
	self.LASTFILTER = nil;
	self.BIDSCROLLFRAME:SortData();
end

-- Update presence list
function sdkp:UpdateRaidPresence(info)
	local nRaidMembers = GetNumRaidMembers();
	
	for raidIndex = 1, nRaidMembers do
		local raidMember = GetRaidRosterInfo(raidIndex);
		local main = self:GetAltInfo(raidMember);
		if main then 
			raidMember = main;	
		end
		
		self.db.profile.raidPresence[raidMember] = true;
	end
end